#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>

#include <exanic/fifo_rx.h>
#include <exanic/fifo_tx.h>
#include <exanic/exanic.h>
#include <exanic/pcie_if.h>

#include "exanic.h"

/* Parses a string of the format "<device>:<port>[:<buffer-size>]" */
static int parse_device_port(const char *str, char *device, int *port_number,
        int *buffer_size)
{
    char buf[20];
    char *tok, *p;

    if (strlen(str) > 16)
        return -1;
    strcpy(buf, str);

    tok = strtok(buf, ":");
    if (tok == NULL)
        return -1;
    strcpy(device, tok);

    tok = strtok(NULL, ":");
    if (tok == NULL)
        return -1;
    *port_number = strtol(tok, &p, 10);
    if (*tok == '\0' || *p != '\0')
        return -1;

    tok = strtok(NULL, ":");
    if (tok == NULL)
        return 0;
    if (strncmp(tok, "0x", 2) != 0)
        return -1;
    *buffer_size = strtol(tok + 2, &p, 16);
    if (*(tok + 2) == '\0' || *p != '\0')
        return -1;

    tok = strtok(NULL, ":");
    if (tok == NULL)
        return 0;

    return -1;
}

int endpoint_open(struct endpoint *e, const char *name, int rx, int tx)
{
    char device[16];
    int port_number;
    int buffer_size = 0x1000;

    memset(e, 0, sizeof(struct endpoint));

    if (parse_device_port(name, device, &port_number, &buffer_size) == -1)
    {
        struct ifreq ifr;

        /* Create socket */
        e->sock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
        if (e->sock == -1)
        {
            fprintf(stderr, "socket failed: %s\n", strerror(errno));
            return -1;
        }

        /* Look up the interface and bind to it */
        memset(&ifr, 0, sizeof(ifr));
        strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name) - 1);
        if (ioctl(e->sock, SIOCGIFINDEX, &ifr) == -1)
        {
            fprintf(stderr, "%s: SIOCGIFINDEX failed: %s\n", ifr.ifr_name,
                    strerror(errno));
            close(e->sock);
            return -1;
        }
        memset(&e->sll, 0, sizeof(e->sll));
        e->sll.sll_family = AF_PACKET;
        e->sll.sll_halen = ETH_ALEN;
        e->sll.sll_ifindex = ifr.ifr_ifindex;

        bind(e->sock, (struct sockaddr *) &e->sll, sizeof(e->sll));

        snprintf(e->name, sizeof(e->name), "%s", name);
        return 0;
    }
    else
    {
        /* Get exanic handle */
        e->exanic = exanic_acquire_handle(device);
        if (e->exanic == NULL)
        {
            fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
            return -1;
        }

        if (rx)
        {
            /* Get exanic rx handle */
            e->rx = exanic_acquire_rx_buffer(e->exanic, port_number, 0);
            if (e->rx == NULL)
            {
                fprintf(stderr, "%s:%d: %s\n", device, port_number,
                        exanic_get_last_error());
                exanic_release_handle(e->exanic);
                return -1;
            }
        }

        if (tx)
        {
            /* Get exanic tx handle */
            e->tx = exanic_acquire_tx_buffer(e->exanic, port_number, buffer_size);
            if (e->tx == NULL)
            {
                fprintf(stderr, "%s:%d: %s\n", device, port_number,
                        exanic_get_last_error());
                if (rx)
                    exanic_release_rx_buffer(e->rx);
                exanic_release_handle(e->exanic);
                return -1;
            }
        }

        snprintf(e->name, sizeof(e->name), "%s:%d", device, port_number);
        return 0;
    }
}

void endpoint_close(struct endpoint *e)
{
    if (e->exanic == NULL)
    {
        close(e->sock);
    }
    else
    {
        if (e->tx)
            exanic_release_tx_buffer(e->tx);
        if (e->rx)
            exanic_release_rx_buffer(e->rx);
        exanic_release_handle(e->exanic);
        e->exanic = NULL;
    }
}

int endpoint_send(struct endpoint *e, const char *buf, size_t len)
{
    if (e->exanic == NULL)
    {
        if (sendto(e->sock, buf, len, 0, (struct sockaddr *) &e->sll,
                sizeof(e->sll)) == -1)
        {
            fprintf(stderr, "%s: sendto failed: %s\n", e->name, strerror(errno));
            return -1;
        }
        return 0;
    }
    else
    {
        if (exanic_transmit_frame(e->tx, buf, len) == -1)
        {
            fprintf(stderr, "%s: %s\n", e->name, exanic_get_last_error());
            return -1;
        }
        return 0;
    }
}

ssize_t endpoint_recv(struct endpoint *e, char *buf, size_t len, int *iter, uint32_t *timestamp)
{
    if (e->exanic == NULL)
    {
        ssize_t size;
        do
            size = recv(e->sock, buf, len, MSG_DONTWAIT);
        while (size == -1 && ++*iter < 10000);
        return size;
    }
    else
    {
        ssize_t size;
        do
            size = exanic_receive_frame(e->rx, buf, len, timestamp);
        while (size <= 0 && ++*iter < 1000);
        return size;
    }
}

inline uint32_t to_exanic_clk_adj(double adj)
{
    if (adj > 0)
        return EXANIC_CLK_ADJ_INC | ((uint32_t)(1 / adj) & 0xFFFFFF);
    else if (adj < 0)
        return EXANIC_CLK_ADJ_DEC | ((uint32_t)(-1 / adj) & 0xFFFFFF);
    else
        return 0;
}

inline double from_exanic_clk_adj(uint32_t adj)
{
    if ((adj & EXANIC_CLK_ADJ_MASK) == 0)
        return 0;
    else if ((adj & EXANIC_CLK_ADJ_INC) != 0)
        return 1.0 / (adj & EXANIC_CLK_ADJ_MASK);
    else
        return -1.0 / (adj & EXANIC_CLK_ADJ_MASK);
}

