#include <math.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/time.h>

#include <exanic/exanic.h>
#include <exanic/pcie_if.h>
#include <exanic/register.h>

#include "common.h"
#include "host.h"

#define LEN 8


struct exanic_host_sync_state
{
    exanic_t *exanic;
    uint32_t tick_hz;   /* Nominal frequency of clock (ticks/second) */
    uint64_t time_tick; /* Time of last measurement (ticks since epoch) */
    double offset_tick; /* Last measured offset of clock (ticks) */
    double adj;         /* Currently applied adjustment value */
    double drift[LEN];  /* Estimated drift of the underlying clock */
    int n;              /* Index into next entry in drift array */
    int startup_mode;   /* Nonzero if measurement history is incomplete */
    int log_next;       /* Make sure next measurement is logged */
    uint64_t last_log_tick; /* Time of last log message (ticks since epoch) */
};


/* Average measurements to estimate underlying clock drift */
static double calc_drift(struct exanic_host_sync_state *state)
{
    double drift;
    int i;

    if (state->startup_mode)
    {
        drift = 0;
        for (i = 0; i < state->n; i++)
            drift += state->drift[i];
        if (state->n > 0)
            drift /= state->n;
    }
    else
    {
        drift = 0;
        for (i = 0; i < LEN; i++)
            drift += state->drift[i];
        drift /= LEN;
    }

    return drift;
}


/* Spin until the next microsecond boundary */
static void gettimeofday_spin(struct timeval *tv)
{
    struct timeval t;
    gettimeofday(&t, NULL);
    do
        gettimeofday(tv, NULL);
    while (t.tv_sec == tv->tv_sec && t.tv_usec == tv->tv_usec);
}


struct exanic_host_sync_state *init_exanic_host_sync(exanic_t *exanic)
{
    struct exanic_host_sync_state *state;
    struct timeval tv;
    uint64_t sys_time_tick;
    uint32_t exanic_time_tick;
    uint32_t exanic_adj;

    state = malloc(sizeof(struct exanic_host_sync_state));

    exanic_retain_handle(exanic);

    state->exanic = exanic;
    state->tick_hz = exanic_register_read(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_HZ));

    log_printf(LOG_INFO, "%s: Starting clock discipline using host clock",
            state->exanic->name);
    log_printf(LOG_INFO, "%s: Nominal frequency: %u Hz",
            state->exanic->name, state->tick_hz);

    /* Set up state struct with current time */
    gettimeofday_spin(&tv);
    exanic_time_tick = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
    sys_time_tick = (uint64_t)tv.tv_sec * state->tick_hz +
            ((uint64_t)tv.tv_usec * state->tick_hz / 1000000);
    state->time_tick = sys_time_tick;
    state->offset_tick = (double)(int32_t)(exanic_time_tick - sys_time_tick);
    state->startup_mode = 1;
    state->log_next = 0;
    state->last_log_tick = sys_time_tick;

    /* Get current adjustment from exanic */
    exanic_adj = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ));
    state->adj = from_exanic_clk_adj(exanic_adj);

    /* Use current adjustment as our initial estimate of drift */
    state->drift[0] = - state->adj;
    state->n = 1;

    return state;
}


enum sync_status poll_exanic_host_sync(struct exanic_host_sync_state *state)
{
    struct timeval tv;
    uint64_t sys_time_tick;
    uint32_t exanic_time_tick;
    double offset_tick, expected_offset_tick;
    double drift, adj;
    uint32_t exanic_adj;
    int fast_poll = 0;
    int i;

    /* Calculate underlying clock drift */
    drift = calc_drift(state);

    /* Do measurements.  This is repeated up to 3 times if we get a
     * measurement that is too far from the expected value. */
    for (i = 0; i < 3; i++)
    {
        /* Get current time from host and from exanic */
        gettimeofday_spin(&tv);
        exanic_time_tick = exanic_register_read(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
        sys_time_tick = (uint64_t)tv.tv_sec * state->tick_hz +
            ((uint64_t)tv.tv_usec * state->tick_hz / 1000000);
        offset_tick = (double)(int32_t)
            (exanic_time_tick - sys_time_tick);

        /* Calculate expected offset of exanic clock based on the
         * current adjustment and estimated drift */
        expected_offset_tick = state->offset_tick +
            (drift + state->adj) * (sys_time_tick - state->time_tick);

        /* Accept measurement if it is within 10us of expected value */
        if (fabs(offset_tick - expected_offset_tick) / state->tick_hz
                < 0.00001)
            break;
    }

    if (i == 3)
    {
        /* Measurement is far from the expected value, so start polling
         * at a faster rate until things stabilise */
        fast_poll = 1;
        state->log_next = 1;
    }

    /* Reset exanic clock if offset is too large (more than 1ms) */
    if (fabs(offset_tick) > 0.001 * state->tick_hz)
    {
        if (!state->startup_mode || state->n > 1)
            log_printf(LOG_WARNING, "%s: Clock error exceeds limits, "
                    "resetting clock: %.0f ticks",
                    state->exanic->name, offset_tick);

        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ), 0);
        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_SET), (uint32_t)sys_time_tick);
        state->time_tick = sys_time_tick;
        state->offset_tick = 0;
        state->adj = 0;
        state->n = 0;
        state->startup_mode = 1;
        state->log_next = 0;
        state->last_log_tick = sys_time_tick;
        return SYNC_FAST_POLL;
    }

    /* Record drift measurement */
    state->drift[state->n] = drift + (offset_tick - expected_offset_tick) /
        (sys_time_tick - state->time_tick);
    if (++state->n >= LEN)
        state->n = state->startup_mode = 0;

    /* Set adjustment to compensate for drift and to correct offset */
    adj = - drift - offset_tick / (state->tick_hz * POLL_INTERVAL);
    exanic_adj = to_exanic_clk_adj(adj);

    exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
            exanic_adj);

    /* Store measurements and current adjustment */
    state->time_tick = sys_time_tick;
    state->offset_tick = offset_tick;
    state->adj = from_exanic_clk_adj(exanic_adj);

    /* Print status */
    if (state->log_next || state->last_log_tick +
            (uint64_t)state->tick_hz * LOG_INTERVAL < sys_time_tick)
    {
        log_printf(LOG_INFO, "%s: Clock offset from host: %.3f us "
                " drift: %.3f ppm", state->exanic->name,
                offset_tick / state->tick_hz * 1000000,
                drift * 1000000);
        state->last_log_tick = sys_time_tick;

        /* Log again if offset is more than 10us */
        state->log_next = (fabs(offset_tick / state->tick_hz) > 0.00001);
    }

    return fast_poll ? SYNC_FAST_POLL : SYNC_OK;
}


void shutdown_exanic_host_sync(struct exanic_host_sync_state *state)
{
    uint32_t exanic_adj;
    double drift;

    log_printf(LOG_INFO, "%s: Stopping clock discipline using host clock",
            state->exanic->name);

    /* Set adjustment to compensate for drift only */
    drift = calc_drift(state);
    exanic_adj = to_exanic_clk_adj(-drift);
    exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
            exanic_adj);

    exanic_release_handle(state->exanic);

    free(state);
}
