#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#include <exanic/exanic.h>
#include <exanic/config.h>
#include <exanic/fifo_rx.h>
#include <exanic/time.h>

#include "pcap-structures.h"

int run = 1;

void signal_handler(int signum)
{
    run = 0;
}

/* Parses a string of the format "<device>:<port>" */
int parse_device_port(const char *str, char *device, int *port_number)
{
    char *p, *q;

    p = strchr(str, ':');
    if (p == NULL)
        return -1;

    if ((p-str) >= 16)
        return -1;
    strncpy(device, str, p - str);
    device[p - str] = '\0';
    *port_number = strtol(p + 1, &q, 10);
    if (*(p + 1) == '\0' || *q != '\0')
        /* strtol failed */
        return -1;
    return 0;
}

void write_pcap_header(FILE *fp, int nsec_pcap)
{
    struct pcap_file_header hdr;
    hdr.magic = nsec_pcap ? NSEC_TCPDUMP_MAGIC : TCPDUMP_MAGIC;
    hdr.version_major = PCAP_VERSION_MAJOR;
    hdr.version_minor = PCAP_VERSION_MINOR;
    hdr.thiszone = 0;
    hdr.sigfigs = 0; /* 9? libpcap always writes 0 */
    hdr.snaplen = 16384;
    hdr.linktype = DLT_EN10MB;
    fwrite(&hdr, sizeof(hdr), 1, fp);
}

void write_pcap_packet(char *data, ssize_t len, struct timespec *ts, int nsec_pcap, FILE *fp)
{
    struct pcap_pkthdr hdr;
    hdr.ts_sec = ts->tv_sec;
    hdr.ts_usec = nsec_pcap ? ts->tv_nsec : (ts->tv_nsec/1000);
    hdr.caplen = len;
    hdr.len = len;
    fwrite(&hdr, sizeof(hdr), 1, fp);
    fwrite(data, 1, len, fp);
}

void print_time(struct timespec *ts)
{
    struct tm tm;
    localtime_r(&ts->tv_sec, &tm);

    printf("%04d%02d%02dT%02d%02d%02d.%09ld ",
           tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
           tm.tm_hour, tm.tm_min, tm.tm_sec,
           ts->tv_nsec);
}

void print_hexdump(char *data, int len)
{
    char ascii[16];
    int i, rem;

    for (i = 0; i < len; )
    {
        if ((i % 16) == 0)
            printf("%04x: ", i);

        printf("%02x", (unsigned char)data[i]);
        if (isprint(data[i]))
            ascii[i%16] = data[i];
        else
            ascii[i%16] = '.';
        i++;

        if ((i % 2) == 0)
            printf(" ");
        if ((i % 16) == 0)
            printf(" %.16s\n", ascii);
    }
    rem = i % 16;
    if (rem)
        printf("%*.*s\n", 2*(16-rem)+((16-rem+1)/2)+rem+1, rem, ascii);
}

int main(int argc, char *argv[])
{
    const char *interface = NULL;
    const char *savefile = NULL;
    FILE *savefp = NULL;
    char device[16];
    int port_number;
    exanic_t *exanic;
    exanic_rx_t *rx;
    char rx_buf[16384];
    ssize_t rx_size;
    uint32_t timestamp;
    struct timespec ts;
    int hw_tstamp = 0, nsec_pcap = 0, flush = 0;
    unsigned long rx_success = 0, rx_aborted = 0, rx_corrupt = 0, rx_hwovfl = 0, rx_swovfl = 0, rx_other = 0;
    int c;

    while ((c = getopt(argc, argv, "i:w:HNh?")) != -1)
    {
        switch (c)
        {
            case 'i':
                interface = optarg;
                break;
            case 'w':
                savefile = optarg;
                break;
            case 'H':
                hw_tstamp = 1;
                break;
            case 'N':
                nsec_pcap = 1;
                break;
            default:
                goto usage_error;
        }
    }

    if (!interface)
        goto usage_error;

    if (exanic_find_port_by_interface_name(interface, device, 16, &port_number) != 0
            && parse_device_port(interface, device, &port_number) != 0)
    {
        fprintf(stderr, "%s: no such interface or not an ExaNIC\n", interface);
        return 1;
    }

    if (savefile)
    {
        if (!strcmp(savefile, "-"))
        {
            savefp = stdout;
            flush = 1;
        }
        else
        {
            savefp = fopen(savefile, "w");
            if (!savefp)
            {
                perror(savefile);
                return 1;
            }
        }
        write_pcap_header(savefp, nsec_pcap);
    }

    /* Get the exanic handle */
    exanic = exanic_acquire_handle(device);
    if (exanic == NULL)
    {
        fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
        goto err_acquire_handle;
    }

    /* Get the exanic rx handle */
    rx = exanic_acquire_rx_buffer(exanic, port_number, 0);
    if (rx == NULL)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number,
                exanic_get_last_error());
        goto err_acquire_rx;
    }

    signal(SIGHUP, signal_handler);
    signal(SIGINT, signal_handler);
    signal(SIGPIPE, signal_handler);
    signal(SIGALRM, signal_handler);
    signal(SIGTERM, signal_handler);

    /* Start reading from the rx buffer */
    while (run)
    {
        rx_size = exanic_receive_frame(rx, rx_buf, sizeof(rx_buf),
                &timestamp);
        if (rx_size > 0)
        {
            if (hw_tstamp)
            {
                uint64_t counter = exanic_timestamp_to_counter(exanic, timestamp);
                ts.tv_sec = counter / 1000000000;
                ts.tv_nsec = counter % 1000000000;
            }
            else
            {
                clock_gettime(CLOCK_REALTIME, &ts);
            }

            if (savefp)
            {
                write_pcap_packet(rx_buf, rx_size, &ts, nsec_pcap, savefp);
                if (flush)
                    fflush(savefp);
            }
            else
            {
                print_time(&ts);
                printf("received %ld bytes\n", rx_size);
                print_hexdump(rx_buf, rx_size);
            }
            rx_success++;
        }
        else if (rx_size < 0)
        {
            clock_gettime(CLOCK_REALTIME, &ts);
            print_time(&ts);
            switch (-rx_size)
            {
                case EXANIC_RX_FRAME_ABORTED:
                    printf("sender aborted frame");
                    rx_aborted++;
                    break;
                case EXANIC_RX_FRAME_CORRUPT:
                    printf("frame with bad CRC");
                    rx_corrupt++;
                    break;
                case EXANIC_RX_FRAME_HWOVFL:
                    printf("frames lost due to insufficient PCIe/memory bandwidth");
                    rx_hwovfl++;
                    break;
                case EXANIC_RX_FRAME_SWOVFL:
                    printf("frames lost due to capture program too slow (usually a scheduling issue)");
                    rx_swovfl++;
                    break;
                case EXANIC_RX_FRAME_TRUNCATED:
                    printf("unexpectedly long (>16KB) frame received");
                    rx_other++;
                    break;
                default:
                    printf("unknown error");
                    rx_other++;
            }
            if ((rx_size != -EXANIC_RX_FRAME_HWOVFL) && (rx_size != -EXANIC_RX_FRAME_SWOVFL))
            {
                printf(", first %d bytes of frame possibly follow:\n", EXANIC_RX_CHUNK_PAYLOAD_SIZE);
                print_hexdump(rx_buf, EXANIC_RX_CHUNK_PAYLOAD_SIZE);
            }
            else
            {
                printf("\n");
            }
        }
    }

    fprintf(stderr, "%s: received=%lu corrupt=%lu aborted=%lu hw_lost=%lu sw_lost=%lu other=%lu\n",
                     argv[0], rx_success, rx_corrupt, rx_aborted, rx_hwovfl, rx_swovfl, rx_other);

    if (savefile)
        fclose(savefp);

    exanic_release_rx_buffer(rx);
    exanic_release_handle(exanic);
    return 0;

err_acquire_rx:
    exanic_release_handle(exanic);
err_acquire_handle:
    return 1;

usage_error:
    fprintf(stderr, "Usage: %s -i interface [-w savefile] [-H] [-N]\n", argv[0]);
    fprintf(stderr, "  -i: specify Linux interface (e.g. eth0) or ExaNIC port name (e.g. exanic0:0)\n");
    fprintf(stderr, "  -w: dump frames to given file in pcap format (- for stdout)\n");
    fprintf(stderr, "  -H: use hardware timestamps (requires exanic-clock-sync or exanic-ptpd)\n");
    fprintf(stderr, "  -N: write nanosecond-resolution pcap format\n");
    return 1;
}
